import numpy

from .._io._helpers import Output

__all__ = [
    "extract",
]


def extract(argv=None):
    import os

    parser = _get_parser()
    args = parser.parse_args(argv)

    # Check that TOUGH output and MESH file exist
    if not os.path.isfile(args.infile):
        raise ValueError("TOUGH output file '{}' not found.".format(args.infile))
    if not os.path.isfile(args.mesh):
        raise ValueError("MESH file '{}' not found.".format(args.mesh))

    # Read MESH and extract X, Y and Z
    nodes, is_eleme = {}, False
    with open(args.mesh, "r") as f:
        for line in f:
            line = line.upper().strip()
            if line[:5].startswith("ELEME"):
                is_eleme = True
                line = next(f)
                while line.strip():
                    label = line[:5]
                    X = float(line[50:60]) if line[50:60].strip() else 0.0
                    Y = float(line[60:70]) if line[60:70].strip() else 0.0
                    Z = float(line[70:80]) if line[70:80].strip() else 0.0
                    nodes[label] = [X, Y, Z]
                    line = next(f)
                headers = ["X", "Y", "Z"]
                break
    if not is_eleme:
        raise ValueError("Invalid MESH file '{}'.".format(args.mesh))

    # Read TOUGH output file
    out = []
    with open(args.infile, "r") as f:
        for line in f:
            line = line.upper().strip()
            if line.startswith("OUTPUT DATA AFTER"):
                out.append(_read_table(f, nodes))

    # Write TOUGH3 element output file
    if not args.split or len(out) == 1:
        with open(args.output_file, "w") as f:
            _write_header(f, headers, out[0])
            for data in out:
                _write_table(f, data, nodes)
    else:
        head, ext = os.path.splitext(args.output_file)
        for i, data in enumerate(out):
            with open("{}_{}{}".format(head, i + 1, ext), "w") as f:
                _write_header(f, headers, data)
                _write_table(f, data, nodes)


def _get_parser():
    import argparse

    # Initialize parser
    parser = argparse.ArgumentParser(
        description=(
            "Extract results from TOUGH main output file and reformat as a TOUGH3 element output file."
        ),
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # Input file
    parser.add_argument(
        "infile", type=str, help="TOUGH output file",
    )

    # Mesh file
    parser.add_argument(
        "mesh", type=str, help="TOUGH MESH file (can be INFILE)",
    )

    # Output file
    parser.add_argument(
        "--output-file",
        "-o",
        type=str,
        default="OUTPUT_ELEME.csv",
        help="TOUGH3 element output file",
    )

    # Split or not
    parser.add_argument(
        "--split",
        "-s",
        default=False,
        action="store_true",
        help="write one file per time step",
    )

    return parser


def _read_table(f, points):
    def str2float(s):
        """Convert variable string to float."""
        try:
            return float(s)
        except ValueError:
            # It's probably something like "0.0001-001"
            significand, exponent = s[:-4], s[-4:]
            return float("{}e{}".format(significand, exponent))

    # Look for "TOTAL TIME"
    while True:
        line = next(f).strip()
        if line.startswith("TOTAL TIME"):
            break

    # Read time step in following line
    line = next(f).strip()
    time = float(line.split()[0])

    # Look for "ELEM."
    while True:
        line = next(f).strip()
        if line.startswith("ELEM."):
            break

    # Read headers once (ignore "ELEM." and "INDEX")
    headers = line.split()[2:]

    # Look for next non-empty line
    while True:
        line = next(f).strip()
        if line:
            break

    # Loop until end of output block
    count = 0
    variables, labels = [], []
    while True:
        if line[:5] in points.keys():
            count += 1
            labels.append(line[:5])
            variables.append([str2float(x) for x in line[5:].split()[1:]])

        line = next(f).strip()
        if line[1:].startswith("@@@@@"):
            break
    if count != len(points):
        raise ValueError("Inconsistent number of elements.")

    return Output(
        time, labels, {k: v for k, v in zip(headers, numpy.transpose(variables))}
    )


def _write_table(f, data, nodes):
    # Write time step
    f.write('"TIME [sec]  {:.8e}"\n'.format(data.time))

    # Loop over elements
    formats = ['"{:>18}"'] + (len(data.data.keys()) + 3) * ["  {:>.12e}"]
    for i, label in enumerate(data.labels):
        record = [label] + nodes[label] + [v[i] for v in data.data.values()]
        record = ",".join(fmt.format(rec) for fmt, rec in zip(formats, record)) + "\n"
        f.write(record)


def _write_header(f, headers, data):
    headers = ["ELEM"] + headers + list(data.data.keys())
    units = [""] + 3 * ["(M)"] + len(data.data.keys()) * ["(-)"]
    f.write(",".join('"{:>18}"'.format(header) for header in headers) + "\n")
    f.write(",".join('"{:>18}"'.format(unit) for unit in units) + "\n")



from __future__ import division, with_statement

import logging

from ._helpers import prune_nones_dict, prune_nones_list, read_record

__all__ = [
    "read",
]


def read(filename):
    """
    Read TOUGH input file.

    Parameters
    ----------
    filename : str
        Input file name.

    Returns
    -------
    dict
        TOUGH input parameters.

    """
    with open(filename, "r") as f:
        out = read_buffer(f)
    return out


def read_buffer(f):
    """Read TOUGH input file."""
    parameters = {}

    # Title
    parameters["title"] = f.readline().strip()

    # Loop over blocks
    while True:
        line = f.readline().strip()

        if line.startswith("ROCKS"):
            rocks, rocks_order = _read_rocks(f)
            parameters.update(rocks)
        elif line.startswith("RPCAP"):
            rpcap = _read_rpcap(f)
            if "default" in parameters.keys():
                parameters["default"].update(rpcap)
            else:
                parameters["default"] = rpcap
        elif line.startswith("FLAC"):
            flac = _read_flac(f, rocks_order)
            parameters["flac"] = flac["flac"]
            for k, v in flac["rocks"].items():
                parameters["rocks"][k].update(v)
        elif line.startswith("MULTI"):
            parameters.update(_read_multi(f))
        elif line.startswith("SELEC"):
            parameters.update(_read_selec(f))
        elif line.startswith("SOLVR"):
            parameters.update(_read_solvr(f))
        elif line.startswith("START"):
            parameters["start"] = True
        elif line.startswith("PARAM"):
            param = _read_param(f)
            parameters["options"] = param["options"]
            parameters["extra_options"] = param["extra_options"]
            if "default" in parameters.keys():
                parameters["default"].update(param["default"])
            else:
                parameters["default"] = param["default"]
        elif line.startswith("INDOM"):
            indom = _read_indom(f)
            for k, v in indom["rocks"].items():
                parameters["rocks"][k].update(v)
        elif line.startswith("MOMOP"):
            parameters.update(_read_momop(f))
        elif line.startswith("TIMES"):
            parameters.update(_read_times(f))
        elif line.startswith("FOFT"):
            parameters.update(_read_oft(f, "element_history"))
        elif line.startswith("COFT"):
            parameters.update(_read_oft(f, "connection_history"))
        elif line.startswith("GOFT"):
            parameters.update(_read_oft(f, "generator_history"))
        elif line.startswith("GENER"):
            parameters.update(_read_gener(f))
        elif line.startswith("DIFFU"):
            parameters.update(_read_diffu(f))
        elif line.startswith("OUTPU"):
            parameters.update(_read_outpu(f))
        elif line.startswith("ELEME"):
            logging.warning("Reading block ELEME is not supported. Skipping.")
        elif line.startswith("CONNE"):
            logging.warning("Reading block CONNE is not supported. Skipping.")
        elif line.startswith("INCON"):
            logging.warning("Reading block INCON is not supported. Skipping.")
        elif line.startswith("NOVER"):
            parameters["nover"] = True
        elif line.startswith("ENDCY"):
            break

    return parameters


def _read_rocks(f):
    """Read ROCKS block data."""
    rocks = {"rocks": {}}
    rocks_order = []

    while True:
        line = f.readline()

        if line.strip():
            # Record 1
            data = read_record(line, "5s,5d,10e,10e,10e,10e,10e,10e,10e")
            rock = data[0]
            rocks["rocks"][rock] = {
                "density": data[2],
                "porosity": data[3],
                "permeability": data[4] if len(set(data[4:7])) == 1 else data[4:7],
                "conductivity": data[7],
                "specific_heat": data[8],
            }

            nad = data[1]
            if nad is not None:
                # Record 2
                line = f.readline()
                data = read_record(line, "10e,10e,10e,10e,10e,10e,10e")
                rocks["rocks"][rock].update(
                    {
                        "compressibility": data[0],
                        "expansion": data[1],
                        "conductivity_dry": data[2],
                        "tortuosity": data[3],
                        "klinkenberg_parameter": data[4],
                        "distribution_coefficient_3": data[5],
                        "distribution_coefficient_4": data[6],
                    }
                )

            if nad and nad > 1:
                rocks["rocks"][rock].update(_read_rpcap(f))

            rocks_order.append(rock)

        else:
            break

    rocks = {
        k: {kk: prune_nones_dict(vv) for kk, vv in v.items()} for k, v in rocks.items()
    }

    return rocks, rocks_order


def _read_rpcap(f):
    """Read RPCAP block data."""
    rpcap = {}

    for key in ["relative_permeability", "capillarity"]:
        line = f.readline()
        data = read_record(line, "5d,5s,10e,10e,10e,10e,10e,10e,10e")
        rpcap[key] = {
            "id": data[0],
            "parameters": prune_nones_list(data[2:]),
        }

    return rpcap


def _read_flac(f, rocks_order):
    """Read FLAC block data."""
    flac = {"rocks": {}, "flac": {}}

    # Record 1
    line = f.readline()
    data = read_record(line, ",".join(16 * ["5d"]))
    flac["flac"]["creep"] = data[0]
    flac["flac"]["porosity_model"] = data[1]
    flac["flac"]["version"] = data[2]

    # Additional records
    for rock in rocks_order:
        flac["rocks"][rock] = {}

        line = f.readline()
        data = read_record(line, "10d,10e,10e,10e,10e,10e,10e,10e")
        flac["rocks"][rock]["permeability_model"] = {
            "id": data[0],
            "parameters": prune_nones_list(data[1:]),
        }

        line = f.readline()
        data = read_record(line, "5d,5s,10e,10e,10e,10e,10e,10e,10e")
        flac["rocks"][rock]["equivalent_pore_pressure"] = {
            "id": data[0],
            "parameters": prune_nones_list(data[2:]),
        }

    flac["flac"] = prune_nones_dict(flac["flac"])

    return flac


def _read_multi(f):
    """Read MULTI block data."""
    multi = {}

    line = f.readline().split()
    multi["n_component"] = int(line[0])
    multi["isothermal"] = int(line[1]) == int(line[0])
    multi["n_phase"] = int(line[2])

    return multi


def _read_selec(f):
    """Read SELEC block data."""
    selec = {"selections": {}}

    line = f.readline()
    data = read_record(line, ",".join(16 * ["5d"]))
    selec["selections"]["integers"] = {k + 1: v for k, v in enumerate(data)}

    if selec["selections"]["integers"][1]:
        selec["selections"]["floats"] = []
        for _ in range(selec["selections"]["integers"][1]):
            line = f.readline()
            data = read_record(line, ",".join(8 * ["10e"]))
            selec["selections"]["floats"] += data

    selec["selections"]["integers"] = prune_nones_dict(selec["selections"]["integers"])
    if selec["selections"]["integers"][1]:
        selec["selections"]["floats"] = prune_nones_list(selec["selections"]["floats"])

    return selec


def _read_solvr(f):
    """Read SOLVR block data."""
    solvr = {}

    line = f.readline()
    data = read_record(line, "1d,2s,2s,3s,2s,10e,10e")
    solvr["solver"] = {
        "method": data[0],
        "z_precond": data[2],
        "o_precond": data[4],
        "rel_iter_max": data[5],
        "eps": data[6],
    }

    return solvr


def _read_param(f):
    """Read PARAM block data."""
    param = {}

    # Record 1
    line = f.readline()
    data = read_record(line, "2d,2d,4d,4d,4d,24S,10s,10e,10e")
    param["options"] = {
        "n_iteration": data[0],
        "verbosity": data[1],
        "n_cycle": data[2],
        "n_second": data[3],
        "n_cycle_print": data[4],
        "temperature_dependence_gas": data[7],
        "effective_strength_vapor": data[8],
    }
    param["extra_options"] = {
        i + 1: int(x) for i, x in enumerate(data[5]) if x.isdigit()
    }

    # Record 2
    line = f.readline()
    data = read_record(line, "10e,10e,10f,10e,10s,10e,10e,10e")
    param["options"].update(
        {
            "t_ini": data[0],
            "t_max": data[1],
            "t_steps": data[2],
            "t_step_max": data[3],
            "gravity": data[5],
            "t_reduce_factor": data[6],
            "mesh_scale_factor": data[7],
        }
    )

    t_steps = int(data[2])
    if t_steps >= 0.0:
        param["options"]["t_steps"] = t_steps
    else:
        param["options"]["t_steps"] = []
        for _ in range(-t_steps):
            line = f.readline()
            data = read_record(line, "10e,10e,10e,10e,10e,10e,10e,10e")
            param["options"]["t_steps"] += prune_nones_list(data)
        if len(param["options"]["t_steps"]) == 1:
            param["options"]["t_steps"] = param["options"]["t_steps"][0]

    # Record 3
    line = f.readline()
    data = read_record(line, "10e,10e,10s,10e,10e,10e")
    param["options"].update(
        {
            "eps1": data[0],
            "eps2": data[1],
            "w_upstream": data[3],
            "w_newton": data[4],
            "derivative_factor": data[5],
        }
    )

    # Record 4
    line = f.readline()
    data = read_record(line, "20e,20e,20e,20e")
    if any(x is not None for x in data):
        data = prune_nones_list(data)
        param["default"] = {"initial_condition": data}
    else:
        param["default"] = {}

    # Remove Nones
    param["options"] = prune_nones_dict(param["options"])
    param["extra_options"] = prune_nones_dict(param["extra_options"])

    return param


def _read_indom(f):
    """Read INDOM block data."""
    indom = {"rocks": {}}

    while True:
        line = f.readline()

        if line.strip():
            rock = line[:5]
            line = f.readline()
            data = read_record(line, "20e,20e,20e,20e")
            data = prune_nones_list(data)
            indom["rocks"][rock] = {"initial_condition": data}
        else:
            break

    return indom


def _read_momop(f):
    """Read MOMOP block data."""
    line = f.readline()
    data = read_record(line, "40S")
    momop = {
        "more_options": {i + 1: int(x) for i, x in enumerate(data[0]) if x.isdigit()}
    }

    return momop


def _read_times(f):
    """Read TIMES block data."""
    times = {"times": []}

    # Record 1
    line = f.readline()
    data = read_record(line, "5d,5d,10e,10e")
    n_times = data[0]

    # Record 2
    while len(times["times"]) < n_times:
        line = f.readline()
        data = read_record(line, "10e,10e,10e,10e,10e,10e,10e,10e")
        times["times"] += prune_nones_list(data)

    if n_times == 1:
        times["times"] = times["times"][0]

    return times


def _read_oft(f, oft):
    """Read FOFT, COFT and GOFT blocks data."""
    history = {oft: []}

    while True:
        line = f.readline().rstrip()

        if line:
            history[oft].append(line)
        else:
            break

    return history


def _read_gener(f):
    """Read GENER block data."""
    gener = {"generators": {}}

    while True:
        line = f.readline()

        if line.strip():
            data = read_record(line, "5s,5s,5d,5d,5d,5d,5s,4s,1s,10e,10e,10e")
            label = data[0]
            tmp = {
                "name": [data[1]],
                "type": [data[7]],
                "layer_thickness": [data[11]],
            }

            ltab = data[5]
            if ltab and ltab > 1:
                for key in ["times", "rates", "specific_enthalpy"]:
                    table = []

                    while len(table) < ltab:
                        line = f.readline()
                        data = read_record(line, "14e,14e,14e,14e")
                        table += prune_nones_list(data)

                    tmp[key] = [table]
            else:
                tmp.update(
                    {
                        "times": [None],
                        "rates": [data[9]],
                        "specific_enthalpy": [data[10]],
                    }
                )

            if label in gener["generators"].keys():
                for k, v in gener["generators"][label].items():
                    v += tmp[k]
            else:
                gener["generators"][label] = tmp
        else:
            break

    # Tidy up
    for generator in gener["generators"].values():
        for k, v in generator.items():
            if len(v) == 1:
                generator[k] = v[0]
            else:
                if all(x is None for x in v):
                    generator[k] = None

    return {
        k: {kk: prune_nones_dict(vv) for kk, vv in v.items()} for k, v in gener.items()
    }


def _read_diffu(f):
    """Read DIFFU block data."""
    diffu = {"diffusion": []}

    for _ in range(2):
        line = f.readline()
        data = read_record(line, "10e,10e,10e,10e,10e,10e,10e,10e")
        diffu["diffusion"].append(prune_nones_list(data))

    return diffu


def _read_outpu(f):
    """Read OUTPU block data."""
    outpu = {"output": {}}

    # Format
    line = f.readline().strip()
    outpu["output"]["format"] = line if line else None

    # Variables
    line = f.readline().strip()
    if line.isdigit():
        n_var = int(line)
        outpu["output"]["variables"] = {}

        for _ in range(n_var):
            line = f.readline()
            data = read_record(line, "20s,5d,5d")
            name = data[0].lower()
            outpu["output"]["variables"][name] = prune_nones_list(data[1:])
            outpu["output"]["variables"][name] = (
                outpu["output"]["variables"][name]
                if len(outpu["output"]["variables"][name]) == 2
                else outpu["output"]["variables"][name][0]
                if len(outpu["output"]["variables"][name]) == 1
                else None
            )

    return outpu



import numpy

from ._structured_grid import structured_grid

__all__ = [
    "voxelize",
]


def voxelize(points, material="dfalt"):
    """
    Generate a 1D hexahedral mesh from 1D point coordinates.

    Parameters
    ----------
    points : array_like (n_points,)
        Cooordinates of points.
    material : str, optional, default 'dfalt'
        Default material name.

    Returns
    -------
    toughio.Mesh
        Output hexahedral mesh.

    """
    if not isinstance(points, (list, tuple, numpy.ndarray)):
        raise TypeError()
    if numpy.ndim(points) != 1:
        raise ValueError()

    points = numpy.asarray(points)
    idx = numpy.argsort(points)
    dx = 0.5 * (points[idx[1:]] - points[idx[:-1]])
    dx = numpy.concatenate(([dx[0]], dx))
    dy = [0.05 * (points.max() - points.min())]
    dz = dy
    origin = -0.5 * numpy.array([dx[0], dy[0], dz[0]])

    mesh = structured_grid(dx, dy, dz, origin)
    cell_type, cells = mesh.cells[0]
    mesh.cells = [(cell_type, cells[idx])]

    return mesh



__all__ = [
    "Parameters",
    "flac",
    "selections",
    "options",
    "extra_options",
    "more_options",
    "solver",
    "generators",
    "output",
    "default",
    "eos",
    "header",
]


_Parameters = {
    "title": "",
    "eos": "",
    "n_component": None,
    "n_phase": None,
    "n_component_mass": None,
    "flac": None,
    "isothermal": False,
    "start": True,
    "nover": False,
    "rocks": {},
    "rocks_order": None,
    "options": {},
    "extra_options": {},
    "more_options": {},
    "selections": {},
    "solver": {},
    "generators": {},
    "diffusion": None,
    "times": None,
    "element_history": None,
    "connection_history": None,
    "generator_history": None,
    "output": None,
    "default": {},
}

Parameters = dict(_Parameters)

flac = {
    "creep": False,
    "porosity_model": None,
    "version": None,
}

selections = {
    "integers": {k + 1: None for k in range(16)},
    "floats": None,
}

options = {
    "n_iteration": None,
    "n_cycle": None,
    "n_second": None,
    "n_cycle_print": None,
    "verbosity": None,
    "temperature_dependence_gas": None,
    "effective_strength_vapor": None,
    "t_ini": None,
    "t_max": None,
    "t_steps": None,
    "t_step_max": None,
    "t_reduce_factor": None,
    "gravity": None,
    "mesh_scale_factor": None,
    "eps1": None,
    "eps2": None,
    "w_upstream": None,
    "w_newton": None,
    "derivative_factor": None,
}

extra_options = {k + 1: None for k in range(24)}

more_options = {k + 1: None for k in range(40)}
more_options[2] = 5

solver = {
    "id": 3,
    "z_precond": "Z0",
    "o_precond": "O0",
    "rel_iter_max": 0.1,
    "eps": 1.0e-6,
}

generators = {
    "name": None,
    "type": None,
    "times": None,
    "rates": None,
    "specific_enthalpy": None,
    "layer_thickness": None,
}

output = {
    "format": "csv",
    "variables": None,
}

default = {
    "density": None,
    "porosity": None,
    "permeability": None,
    "conductivity": None,
    "specific_heat": None,
    "compressibility": None,
    "expansion": None,
    "conductivity_dry": None,
    "tortuosity": None,
    "klinkenberg_parameter": None,
    "distribution_coefficient_3": None,
    "distribution_coefficient_4": None,
    "initial_condition": [None for _ in range(4)],
    "relative_permeability": {"id": None, "parameters": []},
    "capillarity": {"id": None, "parameters": []},
    "permeability_model": {"id": 1, "parameters": []},
    "equivalent_pore_pressure": {"id": 3, "parameters": []},
}

eos = {
    "eos1": [1, 2, 2, 6],
    "eos2": [2, 3, 2, 6],
    "eos3": [2, 3, 2, 6],
    "eos4": [2, 3, 2, 6],
    "eos5": [2, 3, 2, 6],
    "eos7": [3, 3, 2, 6],
    "eos8": [3, 3, 3, 6],
    "eos9": [1, 1, 1, 6],
    "eoswasg": [3, 4, 3, 6],
    "eco2n": [3, 4, 3, 6],
    "eco2n_v2": [3, 4, 3, 6],
    "eco2m": [3, 4, 4, 6],
}

header = "----1----*----2----*----3----*----4----*----5----*----6----*----7----*----8"



import numpy
import pytest

import helpers
import toughio

write_read = lambda x: helpers.write_read(
    "INFILE",
    x,
    toughio.write_input,
    toughio.read_input,
    reader_kws={"file_format": "tough"},
)


def test_title():
    parameters_ref = {"title": helpers.random_string(80)}
    parameters = write_read(parameters_ref)

    assert parameters_ref["title"] == parameters["title"]


def test_rocks():
    keys = [
        "density",
        "porosity",
        "permeability",
        "conductivity",
        "specific_heat",
        "compressibility",
        "expansion",
        "conductivity_dry",
        "tortuosity",
        "klinkenberg_parameter",
        "distribution_coefficient_3",
        "distribution_coefficient_4",
    ]
    parameters_ref = {
        "rocks": {
            helpers.random_string(5): {key: numpy.random.rand() for key in keys[:5]},
            helpers.random_string(5): {
                key: numpy.random.rand()
                if key != "permeability"
                else numpy.random.rand(3)
                for key in keys[:5]
            },
            helpers.random_string(5): {key: numpy.random.rand() for key in keys},
            helpers.random_string(5): {key: numpy.random.rand() for key in keys},
        }
    }
    parameters_ref["rocks"][list(parameters_ref["rocks"].keys())[-1]].update(
        {
            "relative_permeability": {
                "id": numpy.random.randint(10),
                "parameters": numpy.random.rand(numpy.random.randint(7) + 1),
            },
            "capillarity": {
                "id": numpy.random.randint(10),
                "parameters": numpy.random.rand(numpy.random.randint(7) + 1),
            },
        }
    )
    parameters = write_read(parameters_ref)

    assert sorted(parameters_ref["rocks"].keys()) == sorted(parameters["rocks"].keys())

    for k, v in parameters_ref["rocks"].items():
        for kk, vv in v.items():
            if not isinstance(vv, dict):
                assert numpy.allclose(vv, parameters["rocks"][k][kk], atol=1.0e-5)
            else:
                helpers.allclose_dict(vv, parameters["rocks"][k][kk], atol=1.0e-4)


def test_rpcap():
    parameters_ref = {
        "default": {
            "relative_permeability": {
                "id": numpy.random.randint(10),
                "parameters": numpy.random.rand(numpy.random.randint(7) + 1),
            },
            "capillarity": {
                "id": numpy.random.randint(10),
                "parameters": numpy.random.rand(numpy.random.randint(7) + 1),
            },
        },
    }
    parameters = write_read(parameters_ref)

    for k, v in parameters_ref["default"].items():
        helpers.allclose_dict(v, parameters["default"][k], atol=1.0e-4)


def test_flac():
    parameters_ref = {
        "flac": {
            "creep": bool(numpy.random.randint(2)),
            "porosity_model": numpy.random.randint(10),
            "version": numpy.random.randint(10),
        },
        "rocks": {
            helpers.random_string(5): {
                "permeability_model": {
                    "id": numpy.random.randint(10),
                    "parameters": numpy.random.rand(numpy.random.randint(7) + 1),
                },
                "equivalent_pore_pressure": {
                    "id": numpy.random.randint(10),
                    "parameters": numpy.random.rand(numpy.random.randint(7) + 1),
                },
            }
            for _ in numpy.random.rand(10) + 1
        },
    }
    parameters = write_read(parameters_ref)

    helpers.allclose_dict(parameters_ref["flac"], parameters["flac"])
    for k, v in parameters_ref["rocks"].items():
        for kk, vv in v.items():
            helpers.allclose_dict(vv, parameters["rocks"][k][kk], atol=1.0e-4)


@pytest.mark.parametrize("isothermal", [True, False])
def test_multi(isothermal):
    import random
    from toughio._io.tough._common import eos

    parameters_ref = {
        "eos": random.choice(
            [k for k in eos.keys() if k not in {"eos7", "eos8", "eos9"}]
        ),
        "isothermal": isothermal,
    }
    parameters = write_read(parameters_ref)

    multi = [
        parameters["n_component"],
        parameters["n_component"] + 1,
        parameters["n_phase"],
        6,
    ]
    multi_ref = eos[parameters_ref["eos"]]
    assert multi_ref == multi
    assert parameters_ref["isothermal"] == parameters["isothermal"]


def test_selec():
    parameters_ref = {
        "selections": {
            "integers": {
                k + 1: v for k, v in enumerate(numpy.random.randint(100, size=16))
            },
            "floats": numpy.random.rand(numpy.random.randint(100)),
        },
    }
    parameters_ref["selections"]["integers"][1] = (
        len(parameters_ref["selections"]["floats"]) - 1
    ) // 8 + 1
    parameters = write_read(parameters_ref)

    helpers.allclose_dict(
        parameters_ref["selections"]["integers"], parameters["selections"]["integers"]
    )
    if "floats" in parameters["selections"].keys():
        assert numpy.allclose(
            parameters_ref["selections"]["floats"],
            parameters["selections"]["floats"],
            atol=1.0e-4,
        )
    else:
        assert parameters_ref["selections"]["integers"][1] == 0


def test_solvr():
    parameters_ref = {
        "solver": {
            "method": numpy.random.randint(10),
            "z_precond": helpers.random_string(2),
            "o_precond": helpers.random_string(2),
            "rel_iter_max": numpy.random.rand(),
            "eps": numpy.random.rand(),
        },
    }
    parameters = write_read(parameters_ref)

    assert parameters_ref["solver"]["method"] == parameters["solver"]["method"]
    assert parameters_ref["solver"]["z_precond"] == parameters["solver"]["z_precond"]
    assert parameters_ref["solver"]["o_precond"] == parameters["solver"]["o_precond"]
    assert numpy.allclose(
        parameters_ref["solver"]["rel_iter_max"],
        parameters["solver"]["rel_iter_max"],
        atol=1.0e-5,
    )
    assert numpy.allclose(
        parameters_ref["solver"]["eps"], parameters["solver"]["eps"], atol=1.0e-5
    )


@pytest.mark.parametrize(
    "t_steps", [numpy.random.rand(), numpy.random.rand(numpy.random.randint(100) + 1)]
)
def test_param(t_steps):
    parameters_ref = {
        "options": {
            "n_iteration": numpy.random.randint(10),
            "n_cycle": numpy.random.randint(10),
            "n_second": numpy.random.randint(10),
            "n_cycle_print": numpy.random.randint(10),
            "verbosity": numpy.random.randint(10),
            "temperature_dependence_gas": numpy.random.rand(),
            "effective_strength_vapor": numpy.random.rand(),
            "t_ini": numpy.random.rand(),
            "t_max": numpy.random.rand(),
            "t_steps": t_steps,
            "t_step_max": numpy.random.rand(),
            "t_reduce_factor": numpy.random.rand(),
            "gravity": numpy.random.rand(),
            "mesh_scale_factor": numpy.random.rand(),
            "eps1": numpy.random.rand(),
            "eps2": numpy.random.rand(),
            "w_upstream": numpy.random.rand(),
            "w_newton": numpy.random.rand(),
            "derivative_factor": numpy.random.rand(),
        },
        "extra_options": {
            k + 1: v for k, v in enumerate(numpy.random.randint(10, size=24))
        },
        "default": {"initial_condition": numpy.random.rand(numpy.random.randint(5))},
    }
    parameters = write_read(parameters_ref)

    helpers.allclose_dict(parameters_ref["options"], parameters["options"], atol=1.0e-5)
    helpers.allclose_dict(parameters_ref["extra_options"], parameters["extra_options"])
    if "initial_condition" in parameters["default"].keys():
        assert numpy.allclose(
            parameters_ref["default"]["initial_condition"],
            parameters["default"]["initial_condition"],
            atol=1.0e-5,
        )
    else:
        assert not len(parameters_ref["default"]["initial_condition"])


def test_indom():
    parameters_ref = {
        "rocks": {
            helpers.random_string(5): {
                "initial_condition": numpy.random.rand(numpy.random.randint(4) + 1),
            }
            for _ in numpy.random.rand(10) + 1
        },
    }
    parameters = write_read(parameters_ref)

    for k, v in parameters_ref["rocks"].items():
        assert numpy.allclose(
            v["initial_condition"],
            parameters["rocks"][k]["initial_condition"],
            atol=1.0e-4,
        )


def test_momop():
    parameters_ref = {
        "more_options": {
            k + 1: v for k, v in enumerate(numpy.random.randint(10, size=40))
        },
    }
    parameters = write_read(parameters_ref)

    helpers.allclose_dict(parameters_ref["more_options"], parameters["more_options"])


@pytest.mark.parametrize(
    "times", [numpy.random.rand(), numpy.random.rand(numpy.random.randint(100) + 1)]
)
def test_times(times):
    parameters_ref = {"times": times}
    parameters = write_read(parameters_ref)

    assert numpy.allclose(parameters_ref["times"], parameters["times"], atol=1.0e-5)


@pytest.mark.parametrize(
    "oft, n",
    [("element_history", 5), ("connection_history", 10), ("generator_history", 5)],
)
def test_oft(oft, n):
    parameters_ref = {
        oft: [helpers.random_string(n) for _ in range(numpy.random.randint(10) + 1)]
    }
    parameters = write_read(parameters_ref)

    assert parameters_ref[oft] == parameters[oft]


def test_gener():
    n_rnd = numpy.random.randint(100) + 1
    parameters_ref = {
        "generators": {
            helpers.random_string(5): {
                "name": [
                    helpers.random_string(5),
                    helpers.random_string(5),
                    helpers.random_string(5),
                ],
                "type": [
                    helpers.random_string(4),
                    helpers.random_string(4),
                    helpers.random_string(4),
                ],
                "times": [numpy.random.rand(10), None, numpy.random.rand(n_rnd)],
                "rates": [
                    numpy.random.rand(10),
                    numpy.random.rand(),
                    numpy.random.rand(n_rnd),
                ],
                "specific_enthalpy": [
                    numpy.random.rand(10),
                    numpy.random.rand(),
                    numpy.random.rand(n_rnd),
                ],
                "layer_thickness": numpy.random.rand(3),
            },
            helpers.random_string(5): {
                "name": [helpers.random_string(5), helpers.random_string(5)],
                "type": [helpers.random_string(4), helpers.random_string(4)],
                "rates": numpy.random.rand(2),
            },
            helpers.random_string(5): {
                "type": helpers.random_string(4),
                "rates": numpy.random.rand(),
                "layer_thickness": numpy.random.rand(),
            },
        },
    }
    parameters = write_read(parameters_ref)

    assert sorted(parameters_ref["generators"].keys()) == sorted(
        parameters["generators"].keys()
    )

    for k, v in parameters_ref["generators"].items():
        for kk, vv in v.items():
            if kk in {"name", "type"}:
                assert vv == parameters["generators"][k][kk]
            else:
                if numpy.ndim(vv):
                    for i, arr_ref in enumerate(vv):
                        arr = parameters["generators"][k][kk][i]
                        if arr_ref is not None:
                            assert numpy.allclose(arr, arr_ref, atol=1.0e-4)
                        else:
                            assert arr is None
                else:
                    assert numpy.allclose(
                        vv, parameters["generators"][k][kk], atol=1.0e-4
                    )


def test_diffu():
    n_phase = numpy.random.randint(8) + 1
    parameters_ref = {
        "n_phase": n_phase,
        "diffusion": numpy.random.rand(2, n_phase),
    }
    parameters = write_read(parameters_ref)

    assert numpy.allclose(
        parameters_ref["diffusion"], parameters["diffusion"], atol=1.0e-4
    )


def test_outpu():
    parameters_ref = {
        "output": {
            "format": helpers.random_string(20).upper(),
            "variables": {
                helpers.random_string(20): None,
                helpers.random_string(20): [numpy.random.randint(10)],
                helpers.random_string(20): [
                    numpy.random.randint(10),
                    numpy.random.randint(10),
                ],
            },
        },
    }
    parameters = write_read(parameters_ref)

    helpers.allclose_dict(
        parameters_ref["output"]["variables"], parameters["output"]["variables"]
    )


@pytest.mark.parametrize(
    "flag, enable",
    [("start", True), ("start", False), ("nover", True), ("nover", False)],
)
def test_flag(flag, enable):
    parameters_ref = {flag: enable}
    parameters = write_read(parameters_ref)

    if flag in parameters.keys():
        assert parameters_ref[flag] == parameters[flag]
    else:
        assert not enable



from __future__ import with_statement

import collections

import numpy

from . import json, tough

__all__ = [
    "Output",
    "Save",
    "read_input",
    "write_input",
    "read_output",
    "read_save",
]


Output = collections.namedtuple("Output", ["time", "labels", "data"])
Save = collections.namedtuple("Save", ["labels", "data"])


_extension_to_filetype = {
    ".json": "json",
}


def _filetype_from_filename(filename):
    """Determine file type from its extension."""
    import os

    ext = os.path.splitext(filename)[1].lower()
    return (
        _extension_to_filetype[ext] if ext in _extension_to_filetype.keys() else "tough"
    )


def read_input(filename, file_format="tough", **kwargs):
    """
    Read TOUGH input file.

    Parameters
    ----------
    filename : str
        Input file name.
    file_format : str ('tough', 'json'), optional, default 'tough'
        Input file format.

    Returns
    -------
    dict
        TOUGH input parameters.

    """
    if not isinstance(filename, str):
        raise TypeError()
    if file_format not in {"tough", "json"}:
        raise ValueError()

    fmt = file_format if file_format else _filetype_from_filename(filename)

    format_to_reader = {
        "tough": (tough, (), {}),
        "json": (json, (), {}),
    }
    interface, args, default_kwargs = format_to_reader[fmt]
    _kwargs = default_kwargs.copy()
    _kwargs.update(kwargs)
    return interface.read(filename, *args, **_kwargs)


def write_input(filename, parameters, file_format="tough", **kwargs):
    """
    Write TOUGH input file.

    Parameters
    ----------
    filename : str
        Output file name.
    parameters : dict
        Parameters to export.
    file_format : str ('tough', 'json'), optional, default 'tough'
        Output file format.

    """
    if not isinstance(filename, str):
        raise TypeError()
    if not isinstance(parameters, dict):
        raise TypeError()
    if file_format not in {"tough", "json"}:
        raise ValueError()

    fmt = file_format if file_format else _filetype_from_filename(filename)

    format_to_writer = {
        "tough": (tough, (), {}),
        "json": (json, (), {}),
    }
    interface, args, default_kwargs = format_to_writer[fmt]
    _kwargs = default_kwargs.copy()
    _kwargs.update(kwargs)
    interface.write(filename, parameters, *args, **_kwargs)


def read_output(filename, file_format="tough", labels_order=None):
    """
    Read TOUGH output file for each time step.

    Parameters
    ----------
    filename : str
        Input file name.
    file_format : str ('tough' or 'tecplot'), optional, default 'tough'
        TOUGH output file format.
    labels_order : list of array_like or None, optional, default None
        List of labels.

    Returns
    -------
    list of namedtuple
        List of namedtuple (time, labels, data) for each time step.

    """
    if not isinstance(filename, str):
        raise TypeError()
    if file_format not in {"tough", "tecplot"}:
        raise ValueError()
    if not (
        labels_order is None or isinstance(labels_order, (list, tuple, numpy.ndarray))
    ):
        raise TypeError()

    with open(filename, "r") as f:
        if file_format == "tough":
            # Read header
            line = f.readline().replace('"', "")
            headers = [l.strip() for l in line.split(",")]
            headers = headers[1:]

            # Skip second line (unit)
            line = f.readline()

            # Check third line (does it starts with TIME?)
            i = f.tell()
            line = f.readline()
            f.seek(i)
            single = not line.startswith('"TIME')

            # Read data
            if single:
                times, labels, variables = [None], [[]], [[]]
            else:
                times, labels, variables = [], [], []
            line = f.readline().replace('"', "").strip()
            while line:
                line = line.split(",")

                # Time step
                if line[0].startswith("TIME"):
                    line = line[0].split()
                    times.append(float(line[-1]))
                    variables.append([])
                    labels.append([])

                # Output
                else:
                    labels[-1].append(line[0].strip())
                    variables[-1].append([float(l.strip()) for l in line[1:]])

                line = f.readline().strip().replace('"', "")

        elif file_format == "tecplot":
            from ..mesh.tecplot._tecplot import zone_key_to_type

            zone_key_to_type.update({"T": str, "I": int, "J": int})

            # Look for header (VARIABLES)
            while True:
                line = f.readline().strip()
                if line.upper().startswith("VARIABLES"):
                    break

            # Read header (VARIABLES)
            line = line.split("=")[1]
            headers = [l.strip() for l in line.replace('"', "").split()]
            headers = [header for header in headers if "(" not in header]

            # Loop until end of file
            times, labels, variables = [], [], []
            line = f.readline().upper().strip()
            while True:
                # Read zone
                if line.startswith("ZONE"):
                    line = " ".join(line[5:].split())
                    line = "=".join(l.strip() for l in line.split("="))

                    zone = {}
                    i = 0
                    key, value, read_key = "", "", True
                    is_title, is_end = False, False
                    while True:
                        char = line[i]

                        if char == "=":
                            key = key.strip()
                            read_key = False
                            is_title = key == "T"

                        if is_title:
                            # Look for first '"'
                            i += 1
                            while True:
                                if line[i] == '"':
                                    break
                                else:
                                    i += 1

                            # Look for second '"'
                            i += 1
                            while True:
                                value += line[i]
                                if line[i] == '"':
                                    value = value[:-1]
                                    break
                                else:
                                    i += 1
                            is_title, is_end = False, True
                        else:
                            if char != " ":
                                if char != "=":
                                    if read_key:
                                        key += char
                                    else:
                                        value += char
                            else:
                                is_end = True

                        if is_end:
                            if key in zone_key_to_type.keys():
                                zone[key] = zone_key_to_type[key](value)
                            key, value, read_key = "", "", True
                            is_end = False

                        if i >= len(line) - 1:
                            if key in zone_key_to_type.keys():
                                zone[key] = zone_key_to_type[key](value)
                            break
                        else:
                            i += 1

                    zone["T"] = (
                        float(zone["T"].split()[0]) if "T" in zone.keys() else None
                    )
                    if "I" not in zone.keys():
                        raise ValueError()

                # Read data
                data = numpy.genfromtxt(f, max_rows=zone["I"])

                # Output
                times.append(zone["T"])
                labels.append(None)
                variables.append(data)

                line = f.readline().upper().strip()
                if not line:
                    break

    outputs = [
        Output(
            time,
            numpy.array(label),
            {k: v for k, v in zip(headers, numpy.transpose(variable))},
        )
        for time, label, variable in zip(times, labels, variables)
    ]
    return (
        [_reorder_labels(out, labels_order) for out in outputs]
        if labels_order is not None
        else outputs
    )


def read_save(filename, labels_order=None):
    """
    Read TOUGH SAVE file.

    Parameters
    ----------
    filename : str
        Input file name.
    labels_order : list of array_like or None, optional, default None
        List of labels.

    Returns
    -------
    list of namedtuple
        SAVE data as namedtuple (labels, data).

    Note
    ----
    Does not support porosity, USERX and hysteresis values yet.

    """

    def str2float(s):
        """Convert variable string to float."""
        try:
            return float(s)
        except ValueError:
            # It's probably something like "0.0001-001"
            significand, exponent = s[:-4], s[-4:]
            return float("{}e{}".format(significand, exponent))

    if not isinstance(filename, str):
        raise TypeError()
    if not (
        labels_order is None or isinstance(labels_order, (list, tuple, numpy.ndarray))
    ):
        raise TypeError()

    with open(filename, "r") as f:
        # Check first line
        line = f.readline()
        if not line.startswith("INCON"):
            raise ValueError("Invalid SAVE file '{}'.".format(filename))

        # Read data
        labels, variables = [], []
        line = f.readline().strip()
        while line:
            # Label
            line = line.split()
            labels.append(line[0])

            # Primary variables
            line = f.readline().strip().split()
            variables.append([str2float(l) for l in line])

            line = f.readline().strip()
            if line.startswith("+++"):
                break

    output = Save(
        numpy.array(labels),
        {"X{}".format(i + 1): x for i, x in enumerate(numpy.transpose(variables))},
    )
    return _reorder_labels(output, labels_order) if labels_order is not None else output


def _reorder_labels(data, labels):
    """Reorder output or save cell data according to input labels."""
    if len(data.labels) != len(labels):
        raise ValueError()

    mapper = {k: v for v, k in enumerate(data.labels)}
    idx = [mapper[label] for label in labels]
    data.labels[:] = data.labels[idx]

    for k, v in data.data.items():
        data.data[k] = v[idx]

    return data



import numpy

import toughio


def test_extrude_to_3d():
    # Create 2D mesh
    dx = numpy.random.rand(10)
    dy = numpy.random.rand(5)
    mesh_ref = toughio.meshmaker.structured_grid(dx, dy)
    mesh_ref.point_data["points"] = numpy.random.rand(mesh_ref.n_points)
    mesh_ref.cell_data["cells"] = numpy.random.rand(mesh_ref.n_cells)

    # Extrude mesh to 3D
    mesh = mesh_ref.extrude_to_3d(numpy.random.rand(5), axis=2, inplace=False)

    assert mesh.n_points == 6 * mesh_ref.n_points
    assert mesh.n_cells == 5 * mesh_ref.n_cells
    assert mesh.point_data["points"].size == 6 * mesh_ref.n_points
    assert mesh.cell_data["cells"].size == 5 * mesh_ref.n_cells

    for v in mesh.point_data["points"].reshape((6, mesh.n_points // 6)):
        assert numpy.allclose(mesh_ref.point_data["points"], v)

    for v in mesh.cell_data["cells"].reshape((5, mesh.n_cells // 5)):
        assert numpy.allclose(mesh_ref.cell_data["cells"], v)


def test_prune_duplicates():
    # Create mesh with duplicate points and cells
    points = numpy.array(
        [
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [1.0, 1.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
            [1.0, 0.0, 1.0],
            [1.0, 1.0, 1.0],
            [0.0, 1.0, 1.0],
            [2.0, 0.0, 0.0],
            [2.0, 1.0, 0.0],
            [2.0, 0.0, 1.0],
            [2.0, 1.0, 1.0],
            [2.5, 0.5, 0.5],
            [2.5, 0.5, 0.5],
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [1.0, 1.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
            [1.0, 0.0, 1.0],
            [1.0, 1.0, 1.0],
            [0.0, 1.0, 1.0],
        ]
    )
    cells = [
        ("tetra", numpy.array([[8, 12, 9, 10], [10, 13, 9, 11]])),
        (
            "wedge",
            numpy.array(
                [[1, 10, 5, 2, 11, 6], [1, 10, 5, 2, 11, 6], [1, 8, 10, 2, 9, 11]]
            ),
        ),
        (
            "hexahedron",
            numpy.array([[0, 1, 2, 3, 4, 5, 6, 7], [14, 15, 16, 17, 18, 19, 20, 21]]),
        ),
    ]
    mesh = toughio.Mesh(points, cells)
    mesh.point_data["points"] = numpy.random.rand(mesh.n_points)
    mesh.cell_data["cells"] = numpy.random.rand(mesh.n_cells)

    # Remove duplicate points and cells
    mesh.prune_duplicates()

    assert mesh.n_points == 13
    assert mesh.n_cells == 5
    assert mesh.point_data["points"].size == 13
    assert mesh.cell_data["cells"].size == 5



import numpy

import helpers
import toughio


def test_linear():
    helpers.capillarity(
        model=toughio.capillarity.Linear, parameters=[1.0e6, 0.25, 0.4],
    )


def test_pickens():
    helpers.capillarity(
        model=toughio.capillarity.Pickens, parameters=[1.0e6, 0.3, 1.3, 0.8],
    )


def test_trust():
    helpers.capillarity(
        model=toughio.capillarity.TRUST,
        parameters=[1.0e6, 0.3, 1.3, 0.8, 1.0e7],
        sl=numpy.linspace(0.02, 1.0, 50),
    )


def test_milly():
    helpers.capillarity(
        model=toughio.capillarity.Milly, parameters=[0.25],
    )


def test_vanGenuchten():
    helpers.capillarity(
        model=toughio.capillarity.vanGenuchten,
        parameters=[0.457, 0.0, 5.105e-4, 1.0e7, 1.0],
        sl=numpy.linspace(0.02, 1.0, 50),
    )



from __future__ import division, with_statement

import logging
from copy import deepcopy

import numpy

from ._common import default
from ._helpers import (
    add_record,
    block,
    check_parameters,
    dtypes,
    format_data,
    write_multi_record,
    write_record,
)

__all__ = [
    "write",
]


def write(filename, parameters):
    """
    Write TOUGH input file.

    Parameters
    ----------
    filename : str
        Output file name.
    parameters : dict
        Parameters to export.

    """
    from ._common import Parameters, default

    params = deepcopy(Parameters)
    params.update(deepcopy(parameters))

    for k, v in default.items():
        if k not in params["default"].keys():
            params["default"][k] = v

    for rock in params["rocks"].keys():
        for k, v in params["default"].items():
            cond1 = k not in params["rocks"][rock].keys()
            cond2 = k not in {
                "initial_condition",
                "relative_permeability",
                "capillarity",
            }
            if cond1 and cond2:
                params["rocks"][rock][k] = v

    buffer = write_buffer(params)
    with open(filename, "w") as f:
        for record in buffer:
            f.write(record)


@check_parameters(dtypes["PARAMETERS"])
def write_buffer(parameters):
    """Write TOUGH input file as a list of 80-character long record strings."""
    from ._common import eos

    # Check that EOS is defined (for block MULTI)
    if parameters["eos"] and parameters["eos"] not in eos.keys():
        raise ValueError(
            "EOS '{}' is unknown or not supported.".format(parameters["eos"])
        )

    # Set some flags
    cond1 = "relative_permeability" in parameters["default"].keys()
    cond2 = "capillarity" in parameters["default"].keys()
    rpcap = cond1 or cond2

    indom = False
    for rock in parameters["rocks"].values():
        if "initial_condition" in rock.keys():
            if any(x is not None for x in rock["initial_condition"][:4]):
                indom = True
                break

    multi = parameters["eos"] or (parameters["n_component"] and parameters["n_phase"])

    # Check that start is True if indom is True
    if indom and not parameters["start"]:
        logging.warning("Option 'START' is needed to use 'INDOM' conditions.")

    # Define input file contents
    out = ["{:80}\n".format(parameters["title"])]
    out += _write_rocks(parameters)
    out += _write_rpcap(parameters) if rpcap else []
    out += _write_flac(parameters) if parameters["flac"] is not None else []
    out += _write_multi(parameters) if multi else []
    out += _write_selec(parameters) if parameters["selections"] else []
    out += _write_solvr(parameters) if parameters["solver"] else []
    out += _write_start() if parameters["start"] else []
    out += _write_param(parameters)
    out += _write_indom(parameters) if indom else []
    out += _write_momop(parameters) if parameters["more_options"] else []
    out += _write_times(parameters) if parameters["times"] is not None else []
    out += _write_foft(parameters) if parameters["element_history"] is not None else []
    out += (
        _write_coft(parameters) if parameters["connection_history"] is not None else []
    )
    out += (
        _write_goft(parameters) if parameters["generator_history"] is not None else []
    )
    out += _write_gener(parameters) if parameters["generators"] else []
    out += _write_diffu(parameters) if parameters["diffusion"] is not None else []
    out += _write_outpu(parameters) if parameters["output"] else []
    out += _write_nover() if parameters["nover"] else []
    out += _write_endcy()
    return out


@check_parameters(dtypes["ROCKS"], keys="default")
@check_parameters(dtypes["ROCKS"], keys="rocks", is_list=True)
@check_parameters(
    dtypes["MODEL"], keys=("rocks", "relative_permeability"), is_list=True
)
@check_parameters(dtypes["MODEL"], keys=("rocks", "capillarity"), is_list=True)
@block("ROCKS", multi=True)
def _write_rocks(parameters):
    """
    TOUGH input ROCKS block data.

    Introduces material parameters for up to 27 different reservoir domains.

    """
    # Reorder rocks
    if parameters["rocks_order"] is not None:
        order = parameters["rocks_order"]
    else:
        order = parameters["rocks"].keys()

    out = []
    for k in order:
        # Load data
        data = parameters["rocks"][k]

        # Number of additional lines to write per rock
        cond = any(
            data[k] is not None
            for k in [
                "compressibility",
                "expansion",
                "conductivity_dry",
                "tortuosity",
                "klinkenberg_parameter",
                "distribution_coefficient_3",
                "distribution_coefficient_4",
            ]
        )
        nad = (
            2
            if "relative_permeability" in data.keys() or "capillarity" in data.keys()
            else int(cond)
        )

        # Permeability
        per = data["permeability"]
        per = [per] * 3 if not numpy.ndim(per) else per
        if not (isinstance(per, (list, tuple, numpy.ndarray)) and len(per) == 3):
            raise TypeError()

        # Record 1
        out += write_record(
            format_data(
                [
                    (k, "{:5.5}"),
                    (nad if nad else None, "{:>5g}"),
                    (data["density"], "{:>10.4e}"),
                    (data["porosity"], "{:>10.4e}"),
                    (per[0], "{:>10.4e}"),
                    (per[1], "{:>10.4e}"),
                    (per[2], "{:>10.4e}"),
                    (data["conductivity"], "{:>10.4e}"),
                    (data["specific_heat"], "{:>10.4e}"),
                ]
            )
        )

        # Record 2
        if cond:
            out += write_record(
                format_data(
                    [
                        (data["compressibility"], "{:>10.4e}"),
                        (data["expansion"], "{:>10.4e}"),
                        (data["conductivity_dry"], "{:>10.4e}"),
                        (data["tortuosity"], "{:>10.4e}"),
                        (data["klinkenberg_parameter"], "{:>10.4e}"),
                        (data["distribution_coefficient_3"], "{:>10.4e}"),
                        (data["distribution_coefficient_4"], "{:>10.4e}"),
                    ]
                )
            )
        else:
            out += ["{:80}\n".format("")] if nad == 2 else []

        # Relative permeability / Capillary pressure
        if nad == 2:
            out += add_record(data["relative_permeability"])
            out += add_record(data["capillarity"])

    return out


@check_parameters(dtypes["MODEL"], keys=("default", "relative_permeability"))
@check_parameters(dtypes["MODEL"], keys=("default", "capillarity"))
@block("RPCAP")
def _write_rpcap(parameters):
    """
    TOUGH input RPCAP block data (optional).

    Introduces information on relative permeability and capillary pressure functions.

    """
    data = deepcopy(default)
    data.update(parameters["default"])

    out = []
    out += add_record(data["relative_permeability"])
    out += add_record(data["capillarity"])
    return out


@check_parameters(dtypes["FLAC"], keys="flac")
@check_parameters(dtypes["MODEL"], keys=("default", "permeability_model"))
@check_parameters(dtypes["MODEL"], keys=("default", "equivalent_pore_pressure"))
@check_parameters(dtypes["MODEL"], keys=("rocks", "permeability_model"), is_list=True)
@check_parameters(
    dtypes["MODEL"], keys=("rocks", "equivalent_pore_pressure"), is_list=True
)
@block("FLAC", multi=True)
def _write_flac(parameters):
    """
    TOUGH input FLAC block data (optional).

    Introduces mechanical parameters for each material in ROCKS block data.

    """
    # Load data
    from ._common import flac

    data = deepcopy(flac)
    data.update(parameters["flac"])

    # Reorder rocks
    if parameters["rocks_order"]:
        order = parameters["rocks_order"]
    else:
        order = parameters["rocks"].keys()

    # Record 1
    out = write_record(
        format_data(
            [
                (bool(data["creep"]), "{:5g}"),
                (data["porosity_model"], "{:5g}"),
                (data["version"], "{:5g}"),
            ]
        )
    )

    # Additional records
    for k in order:
        # Load data
        data = deepcopy(default)
        data.update(parameters["default"])
        data.update(parameters["rocks"][k])

        # Permeability model
        out += add_record(data["permeability_model"], "{:>10g}")

        # Equivalent pore pressure
        out += add_record(data["equivalent_pore_pressure"])
    return out


@block("MULTI")
def _write_multi(parameters):
    """
    TOUGH input MULTI block (optional).

    Permits the user to select the number and nature of balance equations that will be
    solved. The keyword MULTI is followed by a single data record. For most EOS modules,
    this data block is not needed, as default values are provided internally. Available
    parameter choices are different for different EOS modules.

    """
    from ._common import eos

    out = list(eos[parameters["eos"]]) if parameters["eos"] else [0, 0, 0, 6]
    out[0] = parameters["n_component"] if parameters["n_component"] else out[0]
    out[1] = out[0] if parameters["isothermal"] else out[0] + 1
    out[2] = parameters["n_phase"] if parameters["n_phase"] else out[2]

    # Handle diffusion
    if parameters["diffusion"]:
        out[3] = 8
        parameters["n_phase"] = out[2]  # Save for later check

    # Number of mass components
    if parameters["n_component_mass"]:
        out.append(parameters["n_component_mass"])

    return [("{:>5d}" * len(out) + "\n").format(*out)]


@check_parameters(dtypes["SELEC"], keys="selections")
@block("SELEC")
def _write_selec(parameters):
    """
    TOUGH input SELEC block (optional).

    Introduces a number of integer and floating point parameters that are used for
    different purposes in different TOUGH modules (EOS7, EOS7R, EWASG, T2DM, ECO2N).

    """
    # Load data
    from ._common import selections

    data = deepcopy(selections)
    if parameters["selections"]["integers"]:
        data["integers"].update(parameters["selections"]["integers"])
    if len(parameters["selections"]["floats"]):
        data["floats"] = parameters["selections"]["floats"]

    # Record 1
    out = write_record(
        format_data(
            [(data["integers"][k], "{:>5}") for k in sorted(data["integers"].keys())]
        )
    )

    # Record 2
    if data["floats"] is not None and len(data["floats"]):
        out += write_multi_record(
            format_data([(i, "{:>10.3e}") for i in data["floats"]])
        )
    return out


@check_parameters(dtypes["SOLVR"], keys="solver")
@block("SOLVR")
def _write_solvr(parameters):
    """
    TOUGH input SOLVR block (optional).

    Introduces computation parameters, time stepping information, and default initial
    conditions.

    """
    from ._common import solver

    data = solver.copy()
    data.update(parameters["solver"])
    return write_record(
        format_data(
            [
                (data["method"], "{:1g}  "),
                (data["z_precond"], "{:>2g}   "),
                (data["o_precond"], "{:>2g}"),
                (data["rel_iter_max"], "{:>10.4e}"),
                (data["eps"], "{:>10.4e}"),
            ]
        )
    )


@block("START")
def _write_start():
    """
    TOUGH input START block (optional).

    A record with START typed in columns 1-5 allows a more flexible initialization. More
    specifically, when START is present, INCON data can be in arbitrary order, and need
    not be present for all grid blocks (in which case defaults will be used). Without
    START, there must be a one-to-one correspondence between the data in blocks ELEME
    and INCON.

    """
    from ._common import header

    out = "{:5}{}\n".format("----*", header)
    return [out[:11] + "MOP: 123456789*123456789*1234" + out[40:]]


@check_parameters(dtypes["PARAM"], keys="options")
@check_parameters(dtypes["MOP"], keys="extra_options")
@block("PARAM")
def _write_param(parameters):
    """
    TOUGH input PARAM block data.

    Introduces computation parameters, time stepping information, and default initial
    conditions.

    """
    # Load data
    from ._common import options

    data = options.copy()
    data.update(parameters["options"])

    # Table
    if not isinstance(data["t_steps"], (list, tuple, numpy.ndarray)):
        data["t_steps"] = [data["t_steps"]]

    # Record 1
    from ._common import extra_options

    _mop = deepcopy(extra_options)
    _mop.update(parameters["extra_options"])
    mop = format_data([(_mop[k], "{:>1g}") for k in sorted(_mop.keys())])
    out = write_record(
        format_data(
            [
                (data["n_iteration"], "{:>2g}"),
                (data["verbosity"], "{:>2g}"),
                (data["n_cycle"], "{:>4g}"),
                (data["n_second"], "{:>4g}"),
                (data["n_cycle_print"], "{:>4g}"),
                ("{}".format("".join(mop)), "{:>24}"),
                (None, "{:>10}"),
                (data["temperature_dependence_gas"], "{:>10.4e}"),
                (data["effective_strength_vapor"], "{:>10.4e}"),
            ]
        )
    )

    # Record 2
    out += write_record(
        format_data(
            [
                (data["t_ini"], "{:>10.4e}"),
                (data["t_max"], "{:>10.4e}"),
                (-((len(data["t_steps"]) - 1) // 8 + 1), "{:>9g}."),
                (data["t_step_max"], "{:>10.4e}"),
                (None, "{:>10g}"),
                (data["gravity"], "{:>10.4e}"),
                (data["t_reduce_factor"], "{:>10.4e}"),
                (data["mesh_scale_factor"], "{:>10.4e}"),
            ]
        )
    )

    # Record 2.1
    out += write_multi_record(format_data([(i, "{:>10.4e}") for i in data["t_steps"]]))

    # Record 3
    out += write_record(
        format_data(
            [
                (data["eps1"], "{:>10.4e}"),
                (data["eps2"], "{:>10.4e}"),
                (None, "{:>10.4e}"),
                (data["w_upstream"], "{:>10.4e}"),
                (data["w_newton"], "{:>10.4e}"),
                (data["derivative_factor"], "{:>10.4e}"),
            ]
        )
    )

    # Record 4
    data = parameters["default"]["initial_condition"]
    n = len(data)
    out += write_record(format_data([(i, "{:>20.4e}") for i in data[: min(n, 4)]]))
    return out


@block("INDOM", multi=True)
def _write_indom(parameters):
    """
    TOUGH input INDOM block data (optional).

    Introduces domain-specific initial conditions.

    """
    if parameters["rocks_order"]:
        order = parameters["rocks_order"]
    else:
        order = parameters["rocks"].keys()

    out = []
    for k in order:
        if "initial_condition" in parameters["rocks"][k]:
            data = parameters["rocks"][k]["initial_condition"]
            data = data[: min(len(data), 4)]
            if any(x is not None for x in data):
                out += ["{:5.5}\n".format(k)]
                out += write_record(format_data([(i, "{:>20.4e}") for i in data]))
    return out


@check_parameters(dtypes["MOMOP"], keys="more_options")
@block("MOMOP")
def _write_momop(parameters):
    """
    TOUGH input MOMOP block data (optional).

    Provides additional options.

    """
    from ._common import more_options

    _momop = more_options.copy()
    _momop.update(parameters["more_options"])
    out = write_record(
        format_data([(_momop[k], "{:>1g}") for k in sorted(_momop.keys())])
    )
    return out


@block("TIMES")
def _write_times(parameters):
    """
    TOUGH input TIMES block data (optional).

    Permits the user to obtain printout at specified times.

    """
    data = parameters["times"]
    data = data if numpy.ndim(data) else [data]
    n = len(data)
    out = write_record(format_data([(n, "{:>5g}")]))
    out += write_multi_record(format_data([(i, "{:>10.4e}") for i in data]))
    return out


@block("FOFT", multi=True)
def _write_foft(parameters):
    """
    TOUGH input FOFT block data (optional).

    Introduces a list of elements (grid blocks) for which time-dependent data are to be
    written out for plotting to a file called FOFT during the simulation.

    """
    return write_multi_record(
        format_data([(i, "{:>5g}") for i in parameters["element_history"]]), ncol=1
    )


@block("COFT", multi=True)
def _write_coft(parameters):
    """
    TOUGH input COFT block data (optional).

    Introduces a list of connections for which time-dependent data are to be written out
    for plotting to a file called COFT during the simulation.

    """
    return write_multi_record(
        format_data([(i, "{:>10g}") for i in parameters["connection_history"]]), ncol=1
    )


@block("GOFT", multi=True)
def _write_goft(parameters):
    """
    TOUGH input GOFT block data (optional).

    Introduces a list of sinks/sources for which time-dependent data are to be written
    out for plotting to a file called GOFT during the simulation.

    """
    return write_multi_record(
        format_data([(i, "{:>5g}") for i in parameters["generator_history"]]), ncol=1
    )


@check_parameters(dtypes["GENER"], keys="generators", is_list=True)
@block("GENER", multi=True)
def _write_gener(parameters):
    """
    TOUGH input GENER block data (optional).

    Introduces sinks and/or sources.

    """
    from ._common import generators

    # Handle multicomponent generators
    generator_data = []
    keys = [key for key in generators.keys() if key != "type"]
    for k, v in parameters["generators"].items():
        # Load data
        data = deepcopy(generators)
        data.update(v)

        # Check that data are consistent
        if not isinstance(data["type"], str):
            # Number of components
            num_comps = len(data["type"])

            # Check that values in dict have the same length
            for key in keys:
                if data[key] is not None:
                    if not isinstance(data[key], (list, tuple, numpy.ndarray)):
                        raise TypeError()
                    if len(data[key]) != num_comps:
                        raise ValueError()

            # Split dict
            for i in range(num_comps):
                generator_data.append(
                    (
                        k,
                        {
                            key: (data[key][i] if data[key] is not None else None)
                            for key in generators.keys()
                        },
                    )
                )
        else:
            # Only one component for this element
            # Check that values are scalar or 1D array_like
            for key in keys:
                if numpy.ndim(data[key]) not in {0, 1}:
                    raise ValueError()
            generator_data.append((k, data))

    out = []
    for k, v in generator_data:
        # Table
        ltab, itab = None, None
        if v["times"] is not None and isinstance(
            v["times"], (list, tuple, numpy.ndarray)
        ):
            ltab, itab = len(v["times"]), 1
            if not isinstance(v["rates"], (list, tuple, numpy.ndarray)):
                raise TypeError()
            if not (ltab > 1 and ltab == len(v["rates"])):
                raise ValueError()
        else:
            # Rates and specific enthalpy tables cannot be written without a
            # time table
            for key in ["rates", "specific_enthalpy"]:
                if v[key] is not None:
                    if numpy.ndim(v[key]) != 0:
                        raise ValueError()

        # Record 1
        out += write_record(
            format_data(
                [
                    (k, "{:5.5}"),
                    (v["name"], "{:>5g}"),
                    (None, "{:>5g}"),
                    (None, "{:>5g}"),
                    (None, "{:>5g}"),
                    (ltab, "{:>5g}"),
                    (None, "{:>5g}"),
                    (v["type"], "{:4g}"),
                    (itab, "{:>1g}"),
                    (None if ltab else v["rates"], "{:>10.3e}"),
                    (None if ltab else v["specific_enthalpy"], "{:>10.3e}"),
                    (v["layer_thickness"], "{:>10.3e}"),
                ]
            )
        )

        # Record 2
        if ltab:
            out += write_multi_record(
                format_data([(i, "{:>14.7e}") for i in v["times"]]), ncol=4
            )

        # Record 3
        if ltab:
            out += write_multi_record(
                format_data([(i, "{:>14.7e}") for i in v["rates"]]), ncol=4
            )

        # Record 4
        if ltab and v["specific_enthalpy"] is not None:
            if isinstance(v["specific_enthalpy"], (list, tuple, numpy.ndarray)):
                specific_enthalpy = v["specific_enthalpy"]
            else:
                specific_enthalpy = numpy.full(ltab, v["specific_enthalpy"])
            out += write_multi_record(
                format_data([(i, "{:>14.7e}") for i in specific_enthalpy]), ncol=4
            )
    return out


@block("DIFFU")
def _write_diffu(parameters):
    """
    TOUGH input DIFFU block data (optional).

    Introduces diffusion coefficients.

    """
    if numpy.shape(parameters["diffusion"]) != (2, parameters["n_phase"]):
        raise ValueError()
    mass1, mass2 = parameters["diffusion"]

    out = []
    out += write_multi_record(format_data([(i, "{:>10.3e}") for i in mass1]), ncol=8)
    out += write_multi_record(format_data([(i, "{:>10.3e}") for i in mass2]), ncol=8)
    return out


@check_parameters(dtypes["OUTPU"], keys="output")
@block("OUTPU")
def _write_outpu(parameters):
    """
    TOUGH input OUTPU block data (optional).

    Specifies variables/parameters for printout.

    """
    from ._common import output

    data = deepcopy(output)
    data.update(parameters["output"])

    # Format
    out = []
    out += "{:20}\n".format(data["format"].upper()) if data["format"] else "\n"

    # Variables
    if data["variables"]:
        out += "{:15}\n".format(str(len(data["variables"])))

        for k, v in data["variables"].items():
            fmt = "{:20}"
            if v:
                v = v if isinstance(v, (list, tuple, numpy.ndarray)) else [v]
                if not (0 < len(v) < 3):
                    raise ValueError()
                fmt += "{:5}" * len(v)
                out += "{}\n".format(fmt.format(k.upper(), *v))
            else:
                out += "{}\n".format(fmt.format(k.upper()))

    return out


@block("NOVER")
def _write_nover():
    """TOUGH input NOVER block data (optional)."""
    return []


@block("ENDCY", noend=True)
def _write_endcy():
    """TOUGH input ENDCY block data (optional)."""
    return []



import helpers
import toughio


def test_linear():
    helpers.relative_permeability(
        model=toughio.relative_permeability.Linear, parameters=[0.2, 0.3, 0.8, 0.9],
    )


def test_pickens():
    helpers.relative_permeability(
        model=toughio.relative_permeability.Pickens, parameters=[2.0],
    )


def test_corey():
    helpers.relative_permeability(
        model=toughio.relative_permeability.Corey, parameters=[0.3, 0.05],
    )


def test_grant():
    helpers.relative_permeability(
        model=toughio.relative_permeability.Grant, parameters=[0.3, 0.05],
    )


def test_fatt_klikoff():
    helpers.relative_permeability(
        model=toughio.relative_permeability.FattKlikoff, parameters=[0.3],
    )


def test_vanGenuchten_mualem():
    helpers.relative_permeability(
        model=toughio.relative_permeability.vanGenuchtenMualem,
        parameters=[0.457, 0.15, 1.0, 0.1],
    )


def test_verma():
    helpers.relative_permeability(
        model=toughio.relative_permeability.Verma, parameters=[],
    )



import os

import numpy

import toughio


def test_mesh():
    """Reference values are calculated in FLAC3D."""
    this_dir = os.path.dirname(os.path.abspath(__file__))
    filename = os.path.join(this_dir, "support_files", "all_cell_types.f3grid")
    mesh = toughio.read_mesh(filename)

    volumes = mesh.volumes.sum()
    assert numpy.allclose(volumes, 1.7083333333333344)

    face_areas = sum(face.sum() for face in mesh.face_areas)
    assert numpy.allclose(face_areas, 39.16535341331142)

    face_normals = sum(numpy.abs(face).sum() for face in mesh.face_normals)
    assert numpy.allclose(face_normals, 719.3660094744944)

    connections = mesh.connections.sum()
    assert numpy.allclose(connections, 25353)


def test_cylindric_mesh():
    """Compare volumes and face areas to analytical values."""
    dr = numpy.array([1.0, 2.0, 3.0, 4.0])
    dz = numpy.array([1.0, 2.0, 3.0])
    mesh = toughio.meshmaker.cylindric_grid(dr, dz)

    perimeters = 2.0 * numpy.pi * dr.cumsum()
    base_areas = numpy.pi * dr.cumsum() ** 2
    surface_areas = perimeters * dz.sum()
    section_areas = dr.sum() * dz.sum()

    volumes = mesh.volumes.sum()
    assert numpy.allclose(volumes, base_areas[-1] * dz.sum())

    face_areas_top = numpy.array(mesh.face_areas)[:, 0].sum()
    assert numpy.allclose(face_areas_top, base_areas[-1] * len(dz))

    face_areas_bottom = numpy.array(mesh.face_areas)[:, 1].sum()
    assert numpy.allclose(face_areas_bottom, base_areas[-1] * len(dz))

    face_areas_section_1 = numpy.array(mesh.face_areas)[:, 2].sum()
    assert numpy.allclose(face_areas_section_1, section_areas)

    face_areas_outer = numpy.array(mesh.face_areas)[:, 3].sum()
    assert numpy.allclose(face_areas_outer, surface_areas.sum())

    face_areas_section_2 = numpy.array(mesh.face_areas)[:, 4].sum()
    assert numpy.allclose(face_areas_section_2, section_areas)

    face_areas_inner = numpy.array(mesh.face_areas)[:, 5].sum()
    assert numpy.allclose(face_areas_inner, surface_areas[:-1].sum())



import meshio

from ._helpers import (
    read,
    read_time_series,
    write,
    write_points_cells,
    write_time_series,
)
from ._mesh import Mesh, from_meshio

__all__ = [
    "read",
    "write",
    "write_points_cells",
    "read_time_series",
    "write_time_series",
    "Mesh",
    "from_meshio",
]



from __future__ import division, unicode_literals, with_statement

import logging

import numpy

__all__ = [
    "read",
    "write",
]


def block(keyword):
    """Decorate block writing functions."""

    def decorator(func):
        from functools import wraps

        header = "----1----*----2----*----3----*----4----*----5----*----6----*----7----*----8"

        @wraps(func)
        def wrapper(f, *args):
            f.write("{}{}\n".format(keyword, header))
            func(f, *args)
            f.write("\n")

        return wrapper

    return decorator


def read(filename):
    """
    Read TOUGH MESH file is not supported yet.

    MESH file does not store any geometrical information except node centers.

    """
    raise NotImplementedError("Reading TOUGH MESH file is not supported.")


def write(filename, mesh, nodal_distance, material_name, material_end, incon):
    """Write TOUGH MESH file."""
    if nodal_distance not in {"line", "orthogonal"}:
        raise ValueError()
    if not (material_name is None or isinstance(material_name, dict)):
        raise TypeError()
    if not (material_end is None or isinstance(material_end, (str, list, tuple))):
        raise TypeError()
    if not isinstance(incon, bool):
        raise TypeError()

    # Required variables for blocks ELEME and CONNE
    num_cells = mesh.n_cells
    labels = mesh.labels
    nodes = mesh.centers
    materials = mesh.materials
    volumes = mesh.volumes
    boundary_conditions = (
        mesh.cell_data["boundary_condition"]
        if "boundary_condition" in mesh.cell_data.keys()
        else numpy.zeros(num_cells, dtype=int)
    )
    points = mesh.points
    connections = mesh.connections
    gravity = numpy.array([0.0, 0.0, -1.0])

    # Define parameters related to faces
    faces = mesh.faces
    face_normals = mesh.face_normals
    face_areas = mesh.face_areas

    # Required variables for block INCON
    primary_variables = None
    porosities = None
    permeabilities = None
    if incon:
        primary_variables = (
            mesh.cell_data["initial_condition"]
            if "initial_condition" in mesh.cell_data.keys()
            else primary_variables
        )
        porosities = (
            mesh.cell_data["porosity"]
            if "porosity" in mesh.cell_data.keys()
            else porosities
        )
        permeabilities = (
            mesh.cell_data["permeability"]
            if "permeability" in mesh.cell_data.keys()
            else permeabilities
        )

    # Write MESH and INCON files
    write_buffer(
        filename,
        num_cells,
        labels,
        nodes,
        materials,
        volumes,
        boundary_conditions,
        points,
        connections,
        gravity,
        faces,
        face_normals,
        face_areas,
        primary_variables,
        porosities,
        permeabilities,
        nodal_distance,
        material_name,
        material_end,
        incon,
    )


def write_buffer(
    filename,
    num_cells,
    labels,
    nodes,
    materials,
    volumes,
    boundary_conditions,
    points,
    connections,
    gravity,
    faces,
    face_normals,
    face_areas,
    primary_variables,
    porosities,
    permeabilities,
    nodal_distance,
    material_name,
    material_end,
    incon,
):
    materials = [
        "{:5}".format(material.strip()) if isinstance(material, str) else material
        for material in materials
    ]
    material_name = material_name if material_name else {}
    material_end = material_end if material_end else []
    material_end = [material_end] if isinstance(material_end, str) else material_end

    # Check INCON inputs and show warnings if necessary
    if incon and primary_variables is None:
        logging.warning(
            ("Initial conditions are not defined. " "Writing INCON will be ignored.")
        )

    if porosities is not None:
        if not incon:
            logging.warning("Porosity is only exported if incon is provided.")
        else:
            if not (len(porosities) == num_cells and porosities.ndim == 1):
                raise ValueError("Inconsistent porosity array.")

    if permeabilities is not None:
        if not incon:
            logging.warning(
                "Permeability modifiers are only exported if incon is provided."
            )
        else:
            if permeabilities.ndim not in {1, 2}:
                raise ValueError()
            if not (
                permeabilities.shape == (num_cells, 3)
                if permeabilities.ndim == 2
                else len(permeabilities) == num_cells
            ):
                raise ValueError("Inconsistent permeability modifiers array.")

    # Write MESH file
    with open(filename, "w") as f:
        _write_eleme(
            f,
            labels,
            nodes,
            materials,
            volumes,
            boundary_conditions,
            material_name,
            material_end,
        )
        _write_conne(
            f,
            labels,
            nodes,
            points,
            connections,
            gravity,
            boundary_conditions,
            faces,
            face_normals,
            face_areas,
            nodal_distance,
        )

    # Write INCON file
    if incon and primary_variables is not None:
        import os

        head = os.path.split(filename)[0]
        filename = os.path.join(head, "INCON") if head else "INCON"
        with open(filename, "w") as f:
            _write_incon(
                f, labels, primary_variables, porosities, permeabilities,
            )


@block("ELEME")
def _write_eleme(
    f,
    labels,
    nodes,
    materials,
    volumes,
    boundary_conditions,
    material_name,
    material_end,
):
    """Write ELEME block."""
    # Apply time-independent Dirichlet boundary conditions
    volumes *= numpy.where(boundary_conditions, 1.0e50, 1.0)

    # Write ELEME block
    fmt = "{:5.5}{:>5}{:>5}{:>5}{:10.4e}{:>10}{:>10}{:10.3e}{:10.3e}{:10.3e}\n"
    ending = []
    iterables = zip(labels, materials, volumes, nodes)
    for label, material, volume, node in iterables:
        mat = material_name[material] if material in material_name.keys() else material
        record = fmt.format(
            label,  # ID
            "",  # NSEQ
            "",  # NADD
            mat,  # MAT
            volume,  # VOLX
            "",  # AHTX
            "",  # PMX
            node[0],  # X
            node[1],  # Y
            node[2],  # Z
        )
        if material not in material_end:
            f.write(record)
        else:
            ending.append(record)

    # Append ending cells at the end of the block
    if material_end:
        for record in ending:
            f.write(record)


@block("CONNE")
def _write_conne(
    f,
    labels,
    nodes,
    points,
    connections,
    gravity,
    boundary_conditions,
    faces,
    face_normals,
    face_areas,
    nodal_distance,
):
    """Write CONNE block."""
    # Define unique connection variables
    cell_list = set()
    clabels, centers, int_points, int_normals, areas, bounds = [], [], [], [], [], []
    for i, connection in enumerate(connections):
        if (connection >= 0).any():
            for iface, j in enumerate(connection):
                if j >= 0 and j not in cell_list:
                    # Label
                    clabels.append("{:5.5}{:5.5}".format(labels[i], labels[j]))

                    # Nodal points
                    centers.append([nodes[i], nodes[j]])

                    # Common interface defined by single point and normal vector
                    face = faces[i, iface]
                    int_points.append(points[face[face >= 0][0]])
                    int_normals.append(face_normals[i][iface])

                    # Area of common face
                    areas.append(face_areas[i][iface])

                    # Boundary conditions
                    bounds.append((boundary_conditions[i], boundary_conditions[j]))
        else:
            logging.warning(
                "Element '{}' is not connected to the grid.".format(labels[i])
            )
        cell_list.add(i)

    centers = numpy.array(centers)
    int_points = numpy.array(int_points)
    int_normals = numpy.array(int_normals)
    bounds = numpy.array(bounds)

    # Calculate remaining variables
    lines = numpy.diff(centers, axis=1)[:, 0]
    isot = _isot(lines)
    angles = numpy.dot(lines, gravity) / numpy.linalg.norm(lines, axis=1)

    if nodal_distance == "line":
        fp = _intersection_line_plane(centers[:, 0], lines, int_points, int_normals)
        d1 = numpy.where(
            bounds[:, 0], 1.0e-9, numpy.linalg.norm(centers[:, 0] - fp, axis=1)
        )
        d2 = numpy.where(
            bounds[:, 1], 1.0e-9, numpy.linalg.norm(centers[:, 1] - fp, axis=1)
        )
    elif nodal_distance == "orthogonal":
        d1 = _distance_point_plane(centers[:, 0], int_points, int_normals, bounds[:, 0])
        d2 = _distance_point_plane(centers[:, 1], int_points, int_normals, bounds[:, 1])

    # Write CONNE block
    fmt = "{:10.10}{:>5}{:>5}{:>5}{:>5g}{:10.4e}{:10.4e}{:10.4e}{:10.3e}\n"
    iterables = zip(clabels, isot, d1, d2, areas, angles)
    for label, isot, d1, d2, area, angle in iterables:
        record = fmt.format(
            label,  # ID1-ID2
            "",  # NSEQ
            "",  # NAD1
            "",  # NAD2
            isot,  # ISOT
            d1,  # D1
            d2,  # D2
            area,  # AREAX
            angle,  # BETAX
        )
        f.write(record)


@block("INCON")
def _write_incon(f, labels, primary_variables, porosities, permeabilities):
    """Write INCON block."""
    # Check initial pore pressures
    cond = numpy.logical_and(
        primary_variables[:, 0] > -1.0e9, primary_variables[:, 0] < 0.0,
    )
    if cond.any():
        logging.warning("Negative pore pressures found in 'INCON'.")

    # Write label, porosity and permeability
    buffer = ["{:5.5}".format(label) for label in labels]

    if porosities is not None:
        buffer = [
            buf + "{:10}{:15.9e}".format("", phi)
            for buf, phi in zip(buffer, porosities)
        ]
    else:
        buffer = [buf + "{:10}{:15}".format("", "") for buf in buffer]

    if permeabilities is not None:
        if permeabilities.ndim == 1:
            buffer = [
                buf + "{:10.3e}{:10}{:10}".format(k, "", "")
                for buf, k in zip(buffer, permeabilities)
            ]
        else:
            buffer = [
                buf + "{:10.3e}{:10.3e}{:10.3e}".format(*k)
                for buf, k in zip(buffer, permeabilities)
            ]
    else:
        buffer = [buf + "{:10}{:10}{:10}".format("", "", "") for buf in buffer]

    # Write INCON block
    for pvar, buf in zip(primary_variables, buffer):
        if (pvar > -1.0e9).any():
            f.write(buf + "\n")
            for v in pvar:
                f.write("{:20.4e}".format(v) if v > -1.0e9 else "{:20}".format(""))
            f.write("\n")


def _intersection_line_plane(center, lines, int_points, int_normals):
    """
    Calculate the intersection point between a line and a plane.

    Calculate intersection between a line defined by a point and a direction vector and
    a plane defined by one point and a normal vector.

    """
    tmp = _dot(int_points - center, int_normals) / _dot(lines, int_normals)
    return center + lines * tmp[:, None]


def _distance_point_plane(center, int_points, int_normals, mask):
    """
    Calculate orthogonal distance.

    Calculate orthogonal distance of a point to a plane defined by one point and a
    normal vector.

    """
    return numpy.where(mask, 1.0e-9, numpy.abs(_dot(center - int_points, int_normals)))


def _isot(lines):
    """
    Determine direction of anisotropy.

    Given the direction of the line connecting the cell centers, calculate the
    direction of anisotropy.

    Note
    ----
    It always returns 1 if the connection line is not colinear with X, Y or Z.

    """
    mask = lines != 0.0
    return numpy.where(mask.sum(axis=1) == 1, mask.argmax(axis=1) + 1, 1)


def _dot(A, B):
    """Calculate dot product when arrays A and B have the same shape."""
    return (A * B).sum(axis=1)



from __future__ import print_function

__all__ = [
    "export",
]


format_to_ext = {
    "tecplot": ".dat",
    "vtk": ".vtk",
    "vtu": ".vtu",
    "xdmf": ".xdmf",
}


def export(argv=None):
    import os
    import sys
    from .. import read_mesh, read_output, write_time_series
    from .._io._helpers import _reorder_labels
    from ..meshmaker import voxelize, triangulate

    parser = _get_parser()
    args = parser.parse_args(argv)

    # Display warning if mesh is provided but input file format is 'tecplot'
    # Continue as if mesh was not provided
    with_mesh = bool(args.mesh)
    if args.input_format == "tecplot":
        with_mesh = False
        msg = (
            "Cannot use mesh file with Tecplot TOUGH output, inferring dimensionality"
            if args.mesh
            else "Inferring dimensionality"
        )
    else:
        with_mesh = bool(args.mesh)
        msg = "Mesh file not specified, inferring dimensionality"

    # Check that TOUGH output and mesh file exist
    if not os.path.isfile(args.infile):
        raise ValueError("TOUGH output file '{}' not found.".format(args.infile))
    if with_mesh:
        if not os.path.isfile(args.mesh):
            raise ValueError("Pickled mesh file '{}' not found.".format(args.mesh))

    # Read output file
    print("Reading file '{}' ...".format(args.infile), end="")
    sys.stdout.flush()
    output = read_output(args.infile, args.input_format)
    if args.file_format != "xdmf":
        if args.time_step is not None:
            if not (-len(output) <= args.time_step < len(output)):
                raise ValueError("Inconsistent time step value.")
            output = output[args.time_step]
        else:
            output = output[-1]
    print(" Done!")

    # Triangulate or voxelize if no mesh
    if not with_mesh:
        print("{} ...".format(msg), end="")
        sys.stdout.flush()
        points = _get_points(output if args.file_format != "xdmf" else output[0])
        ndim = 1 if points.ndim == 1 else points.shape[1]
        print(" Done!")

        if ndim > 1:
            print(
                "Mesh is {}D, performing point triangulation ...".format(ndim), end=""
            )
            sys.stdout.flush()

            mesh = triangulate(points)

            if args.file_format != "xdmf":
                for label, data in output.data.items():
                    if label not in {"X", "Y", "Z"}:
                        mesh.add_point_data(label, data)

        else:
            print("Mesh is 1D, voxelizing points ...", end="")
            sys.stdout.flush()

            mesh = voxelize(points)

            if args.file_format != "xdmf":
                for label, data in output.data.items():
                    if label not in {"X", "Y", "Z"}:
                        mesh.add_cell_data(label, data)

    else:
        print("Reading mesh file '{}' ...".format(args.mesh), end="")
        sys.stdout.flush()

        try:
            mesh = read_mesh(args.mesh, file_format="pickle")
        except:
            raise ValueError("Cannot unpickle mesh file {}.".format(args.mesh))

        if args.file_format != "xdmf":
            mesh.read_output(output)
        else:
            output = [_reorder_labels(data, mesh.labels) for data in output]
    print(" Done!")

    # Output file name
    if args.output_file:
        filename = args.output_file
    else:
        head, _ = os.path.splitext(args.infile)
        filename = "{}{}".format(head, format_to_ext[args.file_format])

    # Write output file
    print("Writing output file '{}' ...".format(filename), end="")
    sys.stdout.flush()

    if args.file_format != "xdmf":
        mesh.cell_data.pop("material")
        mesh.write(filename, file_format=args.file_format)

    else:
        data = [
            {k: v for k, v in out.data.items() if k not in {"X", "Y", "Z"}}
            for out in output
        ]
        time_steps = [out.time for out in output]

        if not with_mesh:
            write_time_series(
                filename,
                mesh.points,
                mesh.cells,
                point_data=data,
                time_steps=time_steps,
            )
        else:
            write_time_series(
                filename,
                mesh.points,
                mesh.cells,
                cell_data=data,
                time_steps=time_steps,
            )

    print(" Done!")


def _get_parser():
    import argparse

    # Initialize parser
    parser = argparse.ArgumentParser(
        description=("Export TOUGH simulation results to a file for visualization."),
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # Input file
    parser.add_argument(
        "infile", type=str, help="TOUGH output file",
    )

    # Input file format
    parser.add_argument(
        "--input-format",
        "-i",
        type=str,
        choices=("tough", "tecplot"),
        default="tough",
        help="TOUGH output file format",
    )

    # Mesh file
    parser.add_argument(
        "--mesh", "-m", type=str, default=None, help="pickled toughio.Mesh",
    )

    # Time step
    parser.add_argument(
        "--time-step", "-t", type=int, default=None, help="time step to export",
    )

    # Output file
    parser.add_argument(
        "--output-file", "-o", type=str, default=None, help="exported file",
    )

    # File format
    parser.add_argument(
        "--file-format",
        "-f",
        type=str,
        choices=("tecplot", "vtk", "vtu", "xdmf"),
        default="vtk",
        help="exported file format",
    )

    return parser


def _get_points(output):
    import numpy

    # Number of data points
    n_points = len(next(iter(output.data.values())))

    # Make sure that point coordinates exist
    count = 0

    if "X" in output.data.keys():
        X = numpy.array(output.data["X"])
        count += 1
    else:
        X = numpy.zeros(n_points)

    if "Y" in output.data.keys():
        Y = numpy.array(output.data["Y"])
        count += 1
    else:
        Y = numpy.zeros(n_points)

    if "Z" in output.data.keys():
        Z = numpy.array(output.data["Z"])
        count += 1
    else:
        Z = numpy.zeros(n_points)

    if count == 0:
        raise ValueError("No coordinate array ('X', 'Y', 'Z') found.")

    # Assert dimension of the problem
    nx = len(numpy.unique(X))
    ny = len(numpy.unique(Y))
    nz = len(numpy.unique(Z))

    # Reconstruct points cloud
    points = (
        X
        if ny == 1 and nz == 1
        else Y
        if nx == 1 and nz == 1
        else Z
        if nx == 1 and ny == 1
        else numpy.column_stack((Y, Z))
        if nx == 1
        else numpy.column_stack((X, Z))
        if ny == 1
        else numpy.column_stack((X, Y))
        if nz == 1
        else numpy.column_stack((X, Y, Z))
    )

    return points
